(* $Id: toploop.ml,v 1.11 2021/05/11 02:22:21 garrigue Exp $ *)

open Terms
open Parser
open Eval
open Printer

exception Next of (string list * command * string) Env.t
let next procs = raise (Next procs)

let rec not_seq =
  function
    Seq _ -> false
  | _ -> true

let tracep = ref false

let trace ~env c =
  if !tracep && not_seq c then
    begin
      Format.printf "+ @[%a@]@." print_cmd c;
      let rec loop () =
        Printf.printf "? ";
        flush stdout;
        let q = input_line stdin in
        try
          let (strm__ : _ Stream.t) = lexer (Stream.of_string q) in
          match Stream.peek strm__ with
            Some (Atm "q") -> Stream.junk strm__; raise Sys.Break
          | Some (Ident x) ->
              Stream.junk strm__;
              Format.printf "= %a@." print_val (get ~env x);
              raise Exit
          | Some EOF -> Stream.junk strm__; ()
          | _ -> raise Stream.Failure
        with Exit | Stream.Failure | Stream.Error _ -> loop ()
      in
      loop ()
    end

let () = trace_fun := trace

let cstrm = ref Stream.sempty

let rec toploop ?strm procs =
  try
    let s =
      match strm with
        Some s -> s
      | None ->
          Printf.printf "> ";
          flush stdout;
          let cmd = input_line stdin in
          if cmd = "" then next procs
          else begin cstrm := Stream.of_string cmd; lexer !cstrm end
    in
    let (strm__ : _ Stream.t) = s in
    match Stream.peek strm__ with
      Some (Kwd "define") ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some (Atm f) ->
            Stream.junk strm__;
            let s =
              if strm <> None then s
              else begin cstrm := Stream.of_channel stdin; lexer !cstrm end
            in
            let p = program s in
            Format.printf "= %s defined.@." f; next (Env.add f p procs)
        | _ -> raise (Stream.Error "")
        end
    | Some (Atm "quit") -> Stream.junk strm__; procs
    | Some (Atm "load") ->
        Stream.junk strm__;
        begin match Stream.peek strm__ with
          Some (Atm file) ->
            Stream.junk strm__;
            let file =
              if Sys.file_exists file then file
              else
                let fw = file ^ ".whl" in
                if Sys.file_exists fw then fw
                else
                  let ft = file ^ ".txt" in
                  if Sys.file_exists ft then ft
                  else failwith ("File not found: " ^ file)
            in
            let ic = open_in file in
            cstrm := Stream.of_channel ic;
            let strm = lexer !cstrm in
            let procs = toploop ~strm procs in close_in ic; next procs
        | _ -> raise (Stream.Error "")
        end
    | Some (Atm "trace") ->
        Stream.junk strm__;
        tracep := not !tracep;
        Format.printf "Trace is %s.@." (if !tracep then "on" else "off");
        next procs
    | Some (Atm f) ->
        Stream.junk strm__;
        let exprs =
          try star expr strm__ with Stream.Failure -> raise (Stream.Error "")
        in
        let (l, r) = try unsnoc exprs with _ -> [], Nil in
        let e = build_list r l in
        let r = eval ~procs f (eval_expr ~env:Env.empty e) in
        Format.printf "= %a@." print_val r;
        if strm = None && Stream.next s <> EOF then
          prerr_endline "Garbage at end of line.";
        next procs
    | _ ->
        match try Some (expr strm__) with Stream.Failure -> None with
          Some e ->
            let r = eval_expr ~env:Env.empty e in
            Format.printf "= %a@." print_val r;
            if strm = None && Stream.next s <> EOF then
              prerr_endline "Garbage at end of line.";
            next procs
        | _ ->
            match Stream.peek strm__ with
              Some EOF -> Stream.junk strm__; procs
            | _ -> raise Stream.Failure
  with
    End_of_file -> if strm = None then prerr_endline "Bye."; procs
  | Next procs -> toploop ?strm procs
  | Stream.Failure | Stream.Error "" ->
      let (n, c) = get_pos !cstrm in
      Printf.eprintf "Parsing error before line %d, char %d.\n" n c;
      flush stderr;
      if strm = None then toploop procs else procs
  | Stream.Error s ->
      let (n, c) = get_pos !cstrm in
      Printf.eprintf "Parsing error before line %d, char %d: %s.\n" n c s;
      flush stderr;
      if strm = None then toploop procs else procs
  | exn ->
      prerr_endline (Printexc.to_string exn);
      if strm = None then toploop procs else procs

let start () = Sys.catch_break true; toploop Env.empty

let _ = start ()
